package org.jlab.coda.jevio;

import java.io.*;
import java.nio.*;
import java.nio.channels.FileChannel;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class is used to collect multiple buffers, each containing the bytes
 * of a single evio event (<b>NOT</b> a full evio version 4 formatted buffer).
 * It can write the accumulated events into a properly formatted evio version 4 file.
 * It is theoretically thread-safe.
 * It is designed to be fast and does <b>NOT</b> do a deserialization
 * on the buffers collected.<p>
 *
 * @author timmer
 */
public class EvioCompactEventWriter {

    /** The buffer being written to in evio version 4 format. */
    private ByteBuffer byteBuffer;

    /** Size of the byteBuffer in bytes. */
    private int bufferSize;

    /** Number of buffers flushed to the current file. */
    private int buffersWrittenToFile;

    /**
     * Maximum number of events in a block (events following a block header).
     * Default is {@link EventWriter#DEFAULT_BLOCK_COUNT}.
     */
    private int blockCountMax;

    /** Maximum number of bytes of (non-header) data in a single block. */
    private int blockRoomForData;

    /** Running count of the block number. */
    private int blockNumber = 1;

    /** Position in byteBuffer of the currently-being-used block header. */
    private int currentHeaderPosition;

    /** Size in bytes of the currently-being-used block (includes entire block header). */
    private int currentBlockSize;

    /** Number of events written to the currently-being-used block (not counting dictionary). */
    private int currentBlockEventCount;

    /**
     * Dictionary to include in xml format in the first event of the first block
     * when writing the file.
     */
    private String xmlDictionary;

    /** Number of events written to byteBuffer so far. */
    private int eventsWrittenToBuffer;

    /** Size of block header in bytes. */
    private int headerBytes = 32;

    /** The byte order in which to write a file. */
    private ByteOrder byteOrder;

    /** If true, this object has written it's single file and cannot be reused. */
    private boolean closed;

    //-----------------------
    // File related members
    //-----------------------

    private File currentFile;

    /** The output stream used for writing a file. */
    private FileOutputStream fileOutputStream;

    /** The file channel, used for writing a file, derived from fileOutputStream. */
    private FileChannel fileChannel;

    /** Running count of split output files. */
    private int fileCount;

    /** Part of filename without directory or split numbers. */
    public String baseFileName;

    /** Name of directory in which to place file. */
    private String directory;

    /** Do we split the file into several smaller ones (val > 0)?
     *  If so, this gives the maximum integral number of byteBuffers
     *  to make each file in size. */
    private int split;

    /** Index in baseFileName string where split numbers are to be inserted. */
    private int extensionIndex;

    /** Is it OK to overwrite a previously existing file? */
    private boolean overWriteOK;


    /**
     * This method generates part of a file name given a base file name as an argument.
     * The base file name may contain the characters <b>%d</b> each occurrence of
     * which will be substituted with the given runNumber value. It may also contain
     * characters of the form <b>$(ENV_VAR)</b> which will be substituted with
     * the value of the associated environmental variable or a blank string if
     * none is found.This method also finds any extension and marks its position
     * so if the file is being split, the split number can be inserted immediately
     * before the extension.
     *
     * @param baseName   file name to start with
     * @param runNumber  CODA run number
     * @return           resulting file name after substitutions
     */
    public void generateBaseFileName(String baseName, int runNumber) {

        // Scan for %d which must be replaced by the run number
        baseFileName = baseName.replace("%d", "" + runNumber);

        // Scan for environmental variables of the form $(xxx)
        // and substitute the values for them (blank string if not found)
        if (baseFileName.contains("$(")) {
            Pattern pattern = Pattern.compile("\\$\\((.*?)\\)");
            Matcher matcher = pattern.matcher(baseFileName);
            StringBuffer result = new StringBuffer(100);

            while (matcher.find()) {
                String envVar = matcher.group(1);
                String envVal = System.getenv(envVar);
                if (envVal == null) envVal = "";
//System.out.println("generateFileName: replacing " + envVar + " with " + envVal);
                matcher.appendReplacement(result, envVal);
            }
            matcher.appendTail(result);
//System.out.println("generateFileName: Resulting string = " + result);
            baseFileName = result.toString();
        }

        // Find any existing extension and mark its location
        extensionIndex = baseFileName.lastIndexOf(".");
    }

    /**
     * This method generates a complete file name from the previously determined baseFileName
     * (see {@link #generateBaseFileName(String, int)}).
     * If the file is to be split into multiple smaller files (split > 0), numbers are used to
     * distinguish between the split files starting at 000000 but given by fileCount.
     * These numbers are inserted immediately before any extension or at the end of the file
     * name if no extension is present. The directory string, if any, is prepended to the
     * returned file name.
     *
     * @return generated file name
     */
    public String generateFileName() {
        StringBuilder fileName = new StringBuilder(baseFileName);

        if (split > 0) {
            String splitNum = String.format("%06d", fileCount++);
            if (extensionIndex > 0) {
                fileName.insert(extensionIndex, splitNum);
            }
            else {
                fileName.append(splitNum);
            }
        }

        if (directory != null) {
            fileName.insert(0, "/");
            fileName.insert(0, directory);
        }

        return fileName.toString();
    }


    /**
     * Create an <code>EvioCompactEventWriter</code> object for writing events to a file.
     * It receives buffers containing evio event data (just the event data and <b>not</b> a
     * full evio 4 format) and places them into an internal buffer in full evio 4 format.
     * It writes out the contents of its internal buffer into a file whenever that buffer
     * becomes full. If the split arg is > 0, when the file size reaches the given limit,
     * it is closed and another created. If the file already exists, its contents will be
     * overwritten unless the "overWriteOK" argument is <code>false</code> in
     * which case an exception will be thrown. If the file doesn't exist,
     * it will be created. Byte order defaults to big endian if arg is null.
     * File writing can be terminated by calling {@link #close()} which will flush all
     * remaining data to the file, close it, and disable this object from any more
     * writing.
     *
     * @param baseFilename  base file name used to generate complete file name (may not be null)
     * @param directory     directory in which file is to be placed
     * @param runNumber     number of the CODA run, used in naming files
     * @param split         if < 1, do not split file, write to only one file of unlimited size.
     *                      Else this is max size (in units of integral # of buffers of maxBuffersize
     *                      bytes each) to make a file before closing it and starting writing another.
     * @param bufferSize    number of bytes to make the internal buffer which will
     *                      be storing events before writing them to a file.
     * @param byteOrder     the byte order in which to write the file.
     * @param xmlDictionary dictionary in xml format or null if none.
     *
     * @throws EvioException if baseFileName is null;
     *                       if bufferSize too small;
     *                       if file exists but user requested no over-writing;
     *
     */
    public EvioCompactEventWriter(String baseFilename, String directory, int runNumber, int split,
                                  int bufferSize, ByteOrder byteOrder,
                                  String xmlDictionary)
            throws EvioException {

        // Block size maximum is 1MB or 10k events, whichever is limit is reached first
        this(baseFilename, directory, runNumber, split, 1000000, 10000,
             bufferSize, byteOrder, xmlDictionary, false);
    }


    /**
     * Create an <code>EvioCompactEventWriter</code> object for writing events to a file.
     * It receives buffers containing evio event data (just the event data and <b>not</b> a
     * full evio 4 format) and places them into an internal buffer in full evio 4 format.
     * It writes out the contents of its internal buffer into a file whenever that buffer
     * becomes full. If the split arg is > 0, when the file size reaches the given limit,
     * it is closed and another created. If the file already exists, its contents will be
     * overwritten unless the "overWriteOK" argument is <code>false</code> in
     * which case an exception will be thrown. If the file doesn't exist,
     * it will be created. Byte order defaults to big endian if arg is null.
     * File writing can be terminated by calling {@link #close()} which will flush all
     * remaining data to the file, close it, and disable this object from any more
     * writing.
     *
     * @param baseFilename  base file name used to generate complete file name (may not be null)
     * @param directory     directory in which file is to be placed
     * @param runNumber     number of the CODA run, used in naming files
     * @param split         if < 1, do not split file, write to only one file of unlimited size.
     *                      Else this is max size (in units of integral # of buffers of maxBuffersize
     *                      bytes each) to make a file before closing it and starting writing another.
     * @param blockSizeMax  the max blocksize in bytes which must be >= 400 B and <= 4 MB.
     *                      The size of the block will not be larger than this size
     *                      unless a single event itself is larger.
     * @param blockCountMax the max number of events in a single block which must be
     *                      >= {@link EventWriter#MIN_BLOCK_COUNT} and <= {@link EventWriter#MAX_BLOCK_COUNT}.
     * @param bufferSize    number of bytes to make the internal buffer which will
     *                      be storing events before writing them to a file.
     * @param byteOrder     the byte order in which to write the file.
     * @param xmlDictionary dictionary in xml format or null if none.
     * @param overWriteOK   if <code>false</code> and the file already exists,
     *                      an exception is thrown rather than overwriting it.
     *
     * @throws EvioException if baseFileName is null;
     *                       if blockSizeMax or blockCountMax exceed limits;
     *                       if bufferSize too small;
     *                       if file exists but user requested no over-writing;
     */
    public EvioCompactEventWriter(String baseFilename, String directory, int runNumber, int split,
                                  int blockSizeMax, int blockCountMax, int bufferSize,
                                  ByteOrder byteOrder, String xmlDictionary,
                                  boolean overWriteOK)
            throws EvioException {

        if (baseFilename == null) {
            throw new EvioException("baseFileName arg is null");
        }

        if (blockSizeMax/4 < EventWriter.MIN_BLOCK_SIZE) {
            throw new EvioException("blockSizeMax arg must be bigger or 0");
        }

        if (blockSizeMax/4 > EventWriter.MAX_BLOCK_SIZE) {
            throw new EvioException("blockSizeMax arg must be smaller");
        }

        if (blockCountMax < EventWriter.MIN_BLOCK_COUNT) {
             throw new EvioException("blockCountMax arg must be bigger");
        }

        if (blockCountMax > EventWriter.MAX_BLOCK_COUNT) {
             throw new EvioException("blockCountMax arg must be smaller");
        }

        // Set the size of the internal buffer holding events
        if (bufferSize < 4*18) {
            throw new EvioException("bufferSize arg must be larger (suggest > 10MB)");
        }

        // Store some args permanently
        this.split         = split;
        this.directory     = directory;
        this.bufferSize    = bufferSize;
        this.blockCountMax = blockCountMax;
        this.xmlDictionary = xmlDictionary;
        this.overWriteOK   = overWriteOK;

        // split file suffix starts at 0
        fileCount = 0;

        // Create the file name
        generateBaseFileName(baseFilename, runNumber);
        String fileName = generateFileName();
        currentFile = new File(fileName);

        // If we can't overwrite and file exists, throw exception
        if (!overWriteOK && (currentFile.exists() && currentFile.isFile())) {
            throw new EvioException("File exists but user requested no over-writing, "
                                            + currentFile.getPath());
        }

        // Create internal storage buffer
        byteBuffer = ByteBuffer.allocate(bufferSize);

        // Get byte order straight
        if (byteOrder == null) {
            byteOrder = ByteOrder.BIG_ENDIAN;
        }
        byteBuffer.order(byteOrder);
        this.byteOrder = byteOrder;

        blockNumber = 1;

        // Aim for this amount of data in each block (in bytes)
        blockRoomForData = blockSizeMax - headerBytes;

        // Keep track of how many 32 bit words in current block (after writing header below)
        currentBlockSize = 8;

        // Write out the beginning block header
        // (size & count words are updated when writing event)
        if (xmlDictionary == null) {
            writeHeader(8,0,blockNumber++,false,false,true, false);
        }
        else {
            writeHeader(8,0,blockNumber++,true,false,true,false);
            // Write out any dictionary (currentBlockSize updated)
            writeDictionary();
        }

        // Write out the ending block header so buffer is in proper evio 4 format
        writeEmptyLastBlockHeader(blockNumber);
    }


    /**
     * This routine writes an empty, last block (just the header).
     * This ensures that if the data taking software crashes in the mean time,
     * we still have a file or buffer in the proper format.
     * Make sure that the "last block" bit is set. If we're not closing,
     * the block header simply gets over written in the next write.
     * The position of the buffer or file channel is not changed in preparation
     * for writing over the last block header.
     *
     * @param blockNumber block number to use in header
     * @throws IOException if problems writing to file
     */
    private void writeEmptyLastBlockHeader(int blockNumber)
                    throws EvioException {
        writeHeader(8, 0, blockNumber, false, true, false, true);
    }


    /**
     * Write a block header into the internal buffer.
     *
     * @param words         size in number of 32 bit words
     * @param eventCount    number of events in block
     * @param blockNumber   number of block
     * @param hasDictionary does this block have a dictionary?
     * @param isLast        is this the last block?
     * @param currentHeader should this header be considered the current header?
     * @param absoluteMode  should this header be written so the internal
     *                      buffer's position does <b></b>not</b> change?
     * @throws EvioException if no room in internal buffer to write this block header
     */
    private void writeHeader(int words, int eventCount, int blockNumber,
                             boolean hasDictionary, boolean isLast,
                             boolean currentHeader, boolean absoluteMode) throws EvioException {

        // If no room left for a header to be written ...
        if (byteBuffer.remaining() < 32) {
            throw new EvioException("Undersized internal buffer");
        }

        // Record where beginning of header is so we can
        // go back and update block size and event count.
        if (currentHeader) currentHeaderPosition = byteBuffer.position();

//System.out.println("EventWriter (header): words = " + words +
//        ", block# = " + blockNumber + ", ev Cnt = " + eventCount +
//        ", 6th wd = " + BlockHeaderV4.generateSixthWord(4, hasDictionary, isLast, 0));

    	// Write header words, some of which will be
        // overwritten later when the values are determined.
        if (absoluteMode) {
            int pos =  byteBuffer.position();
            byteBuffer.putInt(pos,      words);       // actual size in 32-bit words (Ints)
            byteBuffer.putInt(pos + 4,  blockNumber); // incremental count of blocks
            byteBuffer.putInt(pos + 8,  8);           // header size always 8
            byteBuffer.putInt(pos + 12, eventCount);  // number of events in block
            byteBuffer.putInt(pos + 16, 0);           // unused
                                                      // version = 4, no event type info
            byteBuffer.putInt(pos + 20, BlockHeaderV4.generateSixthWord(4, hasDictionary, isLast, 0));
            byteBuffer.putInt(pos + 24, 0);           // unused
            byteBuffer.putInt(pos + 28, IBlockHeader.MAGIC_NUMBER); // MAGIC_NUMBER
        }
        else {
            byteBuffer.putInt(words);
            byteBuffer.putInt(blockNumber);
            byteBuffer.putInt(8);
            byteBuffer.putInt(eventCount);
            byteBuffer.putInt(0);
            byteBuffer.putInt(BlockHeaderV4.generateSixthWord(4, hasDictionary, isLast, 0));
            byteBuffer.putInt(0);
            byteBuffer.putInt(IBlockHeader.MAGIC_NUMBER);
        }
    }


    /**
     * Write a dictionary into the internal buffer.
     * @throws EvioException if not enough room in buffer;
     */
    private void writeDictionary() throws EvioException {

        if (xmlDictionary == null || xmlDictionary.length() < 1) {
            throw new EvioException("no dictionary");
        }

        // Turn dictionary data into bytes (characters are endian independent)
        byte[] dictBytes = BaseStructure.stringsToRawBytes(new String[] {xmlDictionary});

        // Write bank header
        byteBuffer.putInt(dictBytes.length/4 + 1);

        if (byteBuffer.order() == ByteOrder.BIG_ENDIAN) {
            byteBuffer.putShort((short)0);
            byteBuffer.put((byte)(DataType.CHARSTAR8.getValue()));
            byteBuffer.put((byte)0);
        }
        else {
            byteBuffer.put((byte)0);
            byteBuffer.put((byte)(DataType.CHARSTAR8.getValue()));
            byteBuffer.putShort((short)0);
        }

        // Write dictionary characters
        byteBuffer.put(dictBytes);

        // Update header's # of 32 bit words written to current block
        // (dictionary does not add to event count).
        currentBlockSize += 2 + dictBytes.length/4;
        byteBuffer.putInt(currentHeaderPosition, currentBlockSize);
	}


    /**
     * Write an event (bank) to the internal buffer in evio version 4 format.
     * The given event buffer must contain only the event's data (event header
     * and event data) and must <b>not</b> be in complete evio format.
     * If the internal buffer is full, it will be flushed to the file.
     *
     * @param eventBuffer the event (bank) to write in buffer form
     * @throws EvioException if event is opposite byte order of internal buffer;
     *                       if close() already called;
     *                       if bad eventBuffer format;
     *                       if file could not be opened for writing;
     *                       if file exists but user requested no over-writing;
     * @throws IOException   if error writing file
     */
    synchronized public void writeEvent(ByteBuffer eventBuffer)
            throws EvioException, IOException {

        if (closed) {
            throw new EvioException("close() has already been called");
        }

        if (eventBuffer.order() != byteOrder) {
            throw new EvioException("event is in wrong byte order");
        }

        // Event size in bytes
        int currentEventBytes = eventBuffer.remaining();

        // Size must be multiple of 4 bytes (whole 32-bit ints)
        if (currentEventBytes % 4 != 0) {
            throw new EvioException("bad eventBuffer format");
        }

        // Number of bytes to be added to internal buffer
        int currentSizeIncrease = currentEventBytes;

        // If we have enough room in the current block and have not exceeded
        // the number of allowed events, or this is the first event,
        // write it in the current block.
        boolean writeNewBlockHeader = true;
//System.out.println("writeEvent: block room, total need " + (currentEventBytes + 4*currentBlockSize - headerBytes)+
//                           ", have " + blockRoomForData);
        if ((((currentEventBytes + 4*currentBlockSize - headerBytes) <= blockRoomForData) &&
                eventsWrittenToBuffer < blockCountMax) || eventsWrittenToBuffer < 1) {
            writeNewBlockHeader = false;
//System.out.println("  Do NOT need another block");
        }
        else {
//System.out.println("  NEED another block");
            // Need more room since now we need another block header
            currentSizeIncrease += headerBytes;
        }

       // If no room in buffer, flush existing buffer to file first.
       // Then write event to newly emptied buffer.
//System.out.println("  room in buffer = " + (bufferSize - byteBuffer.position() - headerBytes) +
//                    ", needed for event = " + currentSizeIncrease);
       if (bufferSize - byteBuffer.position() - headerBytes < currentSizeIncrease) {
//System.out.println("  No room in buffer for this ev, flush to file");
            // Write existing buffer to file, returns true if file split
            if (flushToFile()) {
                // Just after splitting the file, don't write a block header as that is
                // already done in the writeFile() method.
                writeNewBlockHeader = false;
            }
        }
//        else {
//           System.out.println("  Room in buffer for this ev");
//        }

        // Write new block header if required
        if (writeNewBlockHeader) {
            currentBlockSize = 8;
            currentBlockEventCount = 0;
            writeHeader(currentBlockSize, 1, blockNumber++, false, false, true, false);
        }

        // Update the current block header's size and event count
        currentBlockSize += currentEventBytes/4;
        currentBlockEventCount++;
        byteBuffer.putInt(currentHeaderPosition, currentBlockSize);
        byteBuffer.putInt(currentHeaderPosition + EventWriter.EVENT_COUNT_OFFSET,
                          currentBlockEventCount);

        // Write event to internal buffer
        byteBuffer.put(eventBuffer);

        // (Re)Write the last block header so in case of catastrophe, the file is still good
        writeEmptyLastBlockHeader(blockNumber);

        eventsWrittenToBuffer++;
//System.out.println("  writeEvent: events written to buffer = " + eventsWrittenToBuffer);
	}


    /**
     * Flush buffer to file defined in constructor.
     * Does nothing if object already closed.
     *
     * @return <code>true</code>if file was split, else <code>false</code>
     * @throws EvioException if this object already closed;
     *                       if file could not be opened for writing;
     *                       if file exists but user requested no over-writing;
     * @throws IOException   if error writing file
     */
    synchronized public boolean flushToFile() throws EvioException, IOException {
        if (closed) {
            throw new EvioException("close() has already been called");
        }
//System.out.println("flushToFile(): in, write " + eventsWrittenToBuffer + " events");

        // If nothing to write, just return
        if (eventsWrittenToBuffer < 1) {
//System.out.println("flushToFile(): nothing to write");
            return false;
        }

        // In general, the byteBuffer position is always just before the
        // last block header (which is where new events are inserted).
        // Before writing, move the position to the end of the last block header,
        // then flip.
        byteBuffer.position(byteBuffer.position() + headerBytes);

        // Get buffer ready to write
        byteBuffer.flip();

        // This actually creates the file. Do it only once.
        if (buffersWrittenToFile < 1) {
            try {
                fileOutputStream = new FileOutputStream(currentFile, false);  // no appending
                fileChannel = fileOutputStream.getChannel();
            }
            catch (FileNotFoundException e) {
                throw new EvioException("File could not be opened for writing, " +
                                                currentFile.getPath(), e);
            }
        }
        // If appending to existing data, write over last block header (back up 32 bytes)
        else {
            fileChannel.position(fileChannel.position() - 32L);
        }

        // Write everything in internal buffer out to file
        fileChannel.write(byteBuffer);

        // Make sure it writes to the physical file on disk
        fileChannel.force(true);

        // Reset buffer & associated values so it can be reused
        byteBuffer.clear();
        blockNumber = 1;
        eventsWrittenToBuffer = 0;
        currentBlockSize = 8;
        currentBlockEventCount = 0;

        // Write first block header to buffer
        if (xmlDictionary == null) {
            writeHeader(8,0,blockNumber++,false,false,true, false);
        }
        else {
            writeHeader(8,0,blockNumber++,true,false,true,false);
            // Write out any dictionary to the buffer
            writeDictionary();
        }

        // Write last block header to buffer
        writeEmptyLastBlockHeader(blockNumber);

        // When splitting, max file size is <split> integral
        // # of buffers (actual a bit larger).
        // If not splitting, file can be any size.
        buffersWrittenToFile++;

        // If not splitting file, we're done
        if (split < 1 || buffersWrittenToFile < split) {
            return false;
        }

        // Time to split (close this file and start another)

        // Close everything including its associated channel
        fileOutputStream.close();

        // Right now no file is open for writing
        fileOutputStream = null;

        // Create the next file's name
        String fileName = generateFileName();
        currentFile = new File(fileName);

        // If we can't overwrite and file exists, throw exception
        if (!overWriteOK && (currentFile.exists() && currentFile.isFile())) {
            throw new EvioException("File exists but user requested no over-writing, "
                                            + currentFile.getPath());
        }

        buffersWrittenToFile = 0;

        return true;
    }


    /**
     * Get the internal byte buffer. The buffer's position will be 0.
     * It's limit will be the size of the valid data. Basically, it's
     * ready to be read from. The returned buffer shares data with the
     * internal buffer but has separate limit, position, and mark.<p>
     *
     * This will be useful if creating a single buffer to be sent over
     * the network. It probably will <b>not</b> be useful if splitting the file
     * since just after a split the internal buffer contains nothing and
     * in any case will not contain all the data written.
     *
     * @return the internal byte buffer.
     */
    synchronized public ByteBuffer getByteBuffer() {
//System.out.println("getByteBuffer: old buf at pos = " + byteBuffer.position() +
//        ", lim = " + byteBuffer.limit());
        ByteBuffer buf = byteBuffer.duplicate();
//System.out.println("getByteBuffer: duplicate buf at pos = " + buf.position() +
//                ", lim = " + buf.limit() + ", set pos to " + (buf.position() + headerBytes));
        buf.position(buf.position() + headerBytes);
        buf.flip();
        return buf;
    }


    /**
     * This method prints out a portion of a given ByteBuffer object
     * in hex representation of ints.
     *
     * @param buf buffer to be printed out
     * @param lenInInts length of data in ints to be printed
     */
    private static void printBuffer(ByteBuffer buf, int lenInInts) {
        IntBuffer ibuf = buf.asIntBuffer();
        lenInInts = lenInInts > ibuf.capacity() ? ibuf.capacity() : lenInInts;
        for (int i=0; i < lenInInts; i++) {
            System.out.println("  Buf(" + i + ") = 0x" + Integer.toHexString(ibuf.get(i)));
        }
    }


	/** This method flushes any remaining data to file and disables this object. */
    synchronized public void close() {
        if (closed) {
            return;
        }

        // Write any remaining data
        try {
//System.out.println("close(): events written = " + eventsWrittenToBuffer);
            flushToFile();
        }
        catch (EvioException e) {
        }
        catch (IOException e) {
        }

        // Close everything including its associated channel
        try {
            if (fileOutputStream != null) {
                fileOutputStream.close();
            }
        }
        catch (IOException e) {
        }

        closed = true;
    }

}