package org.jlab.coda.jevio;

import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import java.io.*;
import java.nio.*;
import java.nio.channels.FileChannel;
import java.util.ArrayList;

/**
 * This is a class of interest to the user. It is used to read an evio version 4 or earlier
 * format file or buffer. Create an <code>EvioReader</code> from a <code>File</code>
 * object corresponding to an event file, and from this class you can test the file
 * for consistency and, more importantly, you can call {@link #parseNextEvent} or
 * {@link #parseEvent(int)} to get new events and to stream the embedded structures
 * to an IEvioListener. The same can be done by creating an <code>EvioReader</code>
 * from a <code>ByteBuffer</code>.
 *
 * The streaming effect of parsing an event is that the parser will read the event and hand off structures,
 * such as banks, to any IEvioListeners. For those familiar with XML, the event is processed SAX-like.
 * It is up to the listener to decide what to do with the structures.
 * <p>
 * As an alternative to stream processing, after an event is parsed, the user can use the events treeModel
 * for access to the structures. For those familiar with XML, the event is processed DOM-like.
 * <p>
 *
 * @author heddle
 * @author timmer
 *
 */
public class EvioReader {

    /**
	 * This <code>enum</code> denotes the status of a read. <br>
	 * SUCCESS indicates a successful read. <br>
	 * END_OF_FILE indicates that we cannot read because an END_OF_FILE has occurred. Technically this means that what
	 * ever we are trying to read is larger than the buffer's unread bytes.<br>
	 * EVIO_EXCEPTION indicates that an EvioException was thrown during a read, possibly due to out of range values,
	 * such as a negative start position.<br>
	 * UNKNOWN_ERROR indicates that an unrecoverable error has occurred.
	 */
	public static enum ReadStatus {
		SUCCESS, END_OF_FILE, EVIO_EXCEPTION, UNKNOWN_ERROR
	}

	/**
	 * This <code>enum</code> denotes the status of a write.<br>
	 * SUCCESS indicates a successful write. <br>
	 * CANNOT_OPEN_FILE indicates that we cannot write because the destination file cannot be opened.<br>
	 * EVIO_EXCEPTION indicates that an EvioException was thrown during a write.<br>
	 * UNKNOWN_ERROR indicates that an unrecoverable error has occurred.
	 */
	public static enum WriteStatus {
		SUCCESS, CANNOT_OPEN_FILE, EVIO_EXCEPTION, UNKNOWN_ERROR
	}

    /**  Offset to get magic number from start of file. */
    private static final int MAGIC_OFFSET = 28;

    /** Offset to get version number from start of file. */
    private static final int VERSION_OFFSET = 20;

    /** Offset to get block size from start of block. */
    private static final int BLOCK_SIZE_OFFSET = 0;

    /** Mask to get version number from 6th int in block. */
    private static final int VERSION_MASK = 0xff;

    /** Root element tag for XML file */
    private static final String ROOT_ELEMENT = "evio-data";


    /** Skip through a file/buffer and store positions of all the events
     *  in random access mode for versions 4+. */
    private final ArrayList<Long> eventPositions = new ArrayList<Long>(200000);


    /** Used to assign a transient number [1..n] to events as they are being read. */
    private int eventNumber = 0;

    /**
     * This is the number of events in the file. It is not computed unless asked for,
     * and if asked for it is computed and cached in this variable.
     */
    private int eventCount = -1;

    /** Evio version number (1-4). Obtain this by reading first block header. */
    private int evioVersion;

    /**
     * Endianness of the data being read, either
     * {@link java.nio.ByteOrder#BIG_ENDIAN} or
     * {@link java.nio.ByteOrder#LITTLE_ENDIAN}.
     */
    private ByteOrder byteOrder;

    /** Size of the first block in bytes. */
    private int firstBlockSize;

    /**
     * This is the number of blocks in the file including the empty block at the
     * end of the version 4 files. It is not computed unless asked for,
     * and if asked for it is computed and cached in this variable.
     */
    private int blockCount = -1;

	/** The current block header for evio versions 1-3. */
    private BlockHeaderV2 blockHeader2 = new BlockHeaderV2();

    /** The current block header for evio version 4. */
    private BlockHeaderV4 blockHeader4 = new BlockHeaderV4();

    /** Reference to current block header, any version, through interface. */
    private IBlockHeader blockHeader;

    /** Block number expected when reading. Used to check sequence of blocks. */
    private int blockNumberExpected = 1;

    /** If true, throw an exception if block numbers are out of sequence. */
    private boolean checkBlockNumberSequence;

    /** Is this the last block in the file or buffer? */
    private boolean lastBlock;

    /**
     * Version 4 files may have an xml format dictionary in the
     * first event of the first block.
     */
    private String dictionaryXML;

    /** The buffer being read. */
    private ByteBuffer byteBuffer;

    /** Parser object for this file/buffer. */
    private EventParser parser;

    /** Initial position of buffer or mappedByteBuffer when reading a file. */
    private int initialPosition;

    //------------------------
    // File specific members
    //------------------------

    /**
     * The buffer representing a map of the input file which is also
     * accessed through {@link #byteBuffer}.
     */
    private MappedByteBuffer mappedByteBuffer;

    /** The buffer containing an entire block when in sequential reading mode. */
    private ByteBuffer blockBuffer;

    /** Absolute path of the underlying file. */
    private String path;

    /** File object. */
    private File file;

    /** File size in bytes. */
    private long fileSize;

    /** File channel used to read data and access file position. */
    private FileChannel fileChannel;

    /** Data stream used to read data. */
    private DataInputStream dataStream;

    /** Do we need to swap data from file? */
    private boolean swap;

    /**
     * Read this file sequentially and not using a memory mapped buffer.
     * If the file being read > 2.1 GBytes, then this is always true.
     */
    private boolean sequentialRead;

    /** Use this counter so sequential methods can call underlying random access methods. */
    private int sequentialCounter = 1;

    //------------------------
    // EvioReader's state
    //------------------------

    /**
     * This class stores the state of this reader so it can be recovered
     * after a state-changing method has been called -- like {@link #rewind()}.
     */
    private class ReaderState {
        private boolean lastBlock;
        private int eventNumber;
        private long filePosition;
        private int byteBufferPosition;
        private int blockNumberExpected;
        private BlockHeaderV2 blockHeader2;
        private BlockHeaderV4 blockHeader4;
    }

    /**
     * This method saves the current state of this EvioReader object.
     * @return the current state of this EvioReader object.
     */
    private ReaderState getState() {
        ReaderState currentState = new ReaderState();
        currentState.lastBlock   = lastBlock;
        currentState.eventNumber = eventNumber;
        currentState.blockNumberExpected = blockNumberExpected;

        if (sequentialRead) {
            try {
                currentState.filePosition = fileChannel.position();
            }
            catch (IOException e) {/* file is closed so this is irrelevant */}
        }
        else {
            currentState.byteBufferPosition  = byteBuffer.position();
        }

        if (evioVersion > 3) {
            currentState.blockHeader4 = (BlockHeaderV4)blockHeader4.clone();
        }
        else {
            currentState.blockHeader2 = (BlockHeaderV2)blockHeader2.clone();
        }

        return currentState;
    }


    /**
     * This method restores a previously saved state of this EvioReader object.
     * @param state a previously stored state of this EvioReader object.
     */
    private void restoreState(ReaderState state) {
        lastBlock = state.lastBlock;
        eventNumber = state.eventNumber;

        if (sequentialRead) {
            try {
                fileChannel.position(state.filePosition);
            }
            catch (IOException e) {/* file is closed so this is irrelevant */}
        }
        else {
            byteBuffer.position(state.byteBufferPosition);
        }

        blockNumberExpected = state.blockNumberExpected;
        if (evioVersion > 3) {
            blockHeader = blockHeader4 = state.blockHeader4;
        }
        else {
            blockHeader = blockHeader2 = state.blockHeader2;
        }
    }


    /**
     * This method prints out a portion of a given ByteBuffer object
     * in hex representation of ints.
     *
     * @param buf buffer to be printed out
     * @param lenInInts length of data in ints to be printed
     */
    private void printBuffer(ByteBuffer buf, int lenInInts) {
        IntBuffer ibuf = buf.asIntBuffer();
        lenInInts = lenInInts > ibuf.capacity() ? ibuf.capacity() : lenInInts;
        for (int i=0; i < lenInInts; i++) {
            System.out.println("  Buf(" + i + ") = 0x" + Integer.toHexString(ibuf.get(i)));
        }
    }

    //------------------------

    /**
     * Constructor for reading an event file.
     *
     * @param path the full path to the file that contains events.
     *             For writing event files, use an <code>EventWriter</code> object.
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null
     */
    public EvioReader(String path) throws EvioException, IOException {
        this(new File(path));
    }

    /**
     * Constructor for reading an event file.
     *
     * @param path the full path to the file that contains events.
     *             For writing event files, use an <code>EventWriter</code> object.
     * @param checkBlkNumSeq if <code>true</code> check the block number sequence
     *                       and throw an exception if it is not sequential starting
     *                       with 1
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public EvioReader(String path, boolean checkBlkNumSeq) throws EvioException, IOException {
        this(new File(path), checkBlkNumSeq);
    }

    /**
     * Constructor for reading an event file.
     *
     * @param file the file that contains events.
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null
     */
    public EvioReader(File file) throws EvioException, IOException {
        this(file, false);
    }


    /**
     * Constructor for reading an event file.
     *
     * @param file the file that contains events.
     * @param checkBlkNumSeq if <code>true</code> check the block number sequence
     *                       and throw an exception if it is not sequential starting
     *                       with 1
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public EvioReader(File file, boolean checkBlkNumSeq)
                                        throws EvioException, IOException {
        this(file, false, false);
    }


    /**
     * Constructor for reading an event file.
     *
     * @param path the full path to the file that contains events.
     *             For writing event files, use an <code>EventWriter</code> object.
     * @param checkBlkNumSeq if <code>true</code> check the block number sequence
     *                       and throw an exception if it is not sequential starting
     *                       with 1
     * @param sequential     if <code>true</code> read the file sequentially and not
     *                       using a memory mapped buffer. If file > 2.1 GB, then reads
     *                       are always sequential
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public EvioReader(String path, boolean checkBlkNumSeq, boolean sequential)
            throws EvioException, IOException {
        this(new File(path), checkBlkNumSeq, sequential);
    }


    /**
     * Constructor for reading an event file.
     *
     * @param file the file that contains events.
     * @param checkBlkNumSeq if <code>true</code> check the block number sequence
     *                       and throw an exception if it is not sequential starting
     *                       with 1
     * @param sequential     if <code>true</code> read the file sequentially and not
     *                       using a memory mapped buffer. If file > 2.1 GB, then reads
     *                       are always sequential
     *
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public EvioReader(File file, boolean checkBlkNumSeq, boolean sequential)
                                        throws EvioException, IOException {
        if (file == null) {
            throw new EvioException("File arg is null");
        }
        this.file = file;

        checkBlockNumberSequence = checkBlkNumSeq;
        initialPosition = 0;

        FileInputStream fileInputStream = new FileInputStream(file);
        path = file.getAbsolutePath();
        fileChannel = fileInputStream.getChannel();
        fileSize = fileChannel.size();

        // Is the file byte size > max int value?
        // If so we cannot use a memory mapped file.
        sequentialRead = sequential;
        if (fileSize > Integer.MAX_VALUE) {
            sequentialRead = true;
        }

        if (sequentialRead) {
            dataStream = new DataInputStream(fileInputStream);
//System.out.println("Big File or Reading Sequentially");
        }
        else {
            mapFile(fileChannel);
            fileChannel.close(); // this object is no longer needed since we have the map
//System.out.println("Little file or using Memory Map");
        }

        // Read first block header and find the file's endianness & evio version #.
        // If there's a dictionary, read that too.
        if (getFirstHeader() != ReadStatus.SUCCESS) {
            throw new IOException("Failed reading first block header/dictionary");
        }

        // For the latest evio format, generate a table
        // of all event positions in buffer for random access.
        if (evioVersion > 3) {
            generateEventPositionTable();
        }

        parser = new EventParser();
    }


    /**
     * Constructor for reading a buffer.
     *
     * @param byteBuffer the buffer that contains events.
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null
     */
    public EvioReader(ByteBuffer byteBuffer) throws EvioException, IOException {
        this(byteBuffer, false);
    }

    /**
     * Constructor for reading a buffer.
     *
     * @param byteBuffer the buffer that contains events.
     * @param checkBlkNumSeq if <code>true</code> check the block number sequence
     *                       and throw an exception if it is not sequential starting
     *                       with 1
     * @see EventWriter
     * @throws IOException   if read failure
     * @throws EvioException if file arg is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public EvioReader(ByteBuffer byteBuffer, boolean checkBlkNumSeq)
                                                    throws EvioException, IOException {

        if (byteBuffer == null) {
            throw new EvioException("Buffer arg is null");
        }

        checkBlockNumberSequence = checkBlkNumSeq;
        initialPosition = byteBuffer.position();
        this.byteBuffer = byteBuffer;

        // Read first block header and find the file's endianness & evio version #.
        // If there's a dictionary, read that too.
        if (getFirstHeader() != ReadStatus.SUCCESS) {
            throw new IOException("Failed reading first block header/dictionary");
        }

        // For the latest evio format, generate a table
        // of all event positions in buffer for random access.
        if (evioVersion > 3) {
            generateEventPositionTable();
        }

        parser = new EventParser();
    }

    /**
     * This method can be used to avoid creating additional EvioReader
     * objects by reusing this one with another buffer. The method
     * {@link #close()} is called before anything else.
     *
     * @param buf ByteBuffer to be read
     * @throws IOException   if read failure
     * @throws EvioException if buf is null;
     *                       if first block number != 1 when checkBlkNumSeq arg is true
     */
    public synchronized void setBuffer(ByteBuffer buf) throws EvioException, IOException {

        if (buf == null) {
            throw new EvioException("arg is null");
        }

        close();

        lastBlock           = false;
        eventNumber         = 0;
        blockCount          = -1;
        eventCount          = -1;
        blockNumberExpected = 1;
        sequentialCounter   = 1;
        dictionaryXML       = null;
        initialPosition     = buf.position();
        byteBuffer          = buf;

        if (getFirstHeader() != ReadStatus.SUCCESS) {
            throw new IOException("Failed reading first block header/dictionary");
        }

        if (evioVersion > 3) {
            generateEventPositionTable();
        }
    }

    /**
     * Is this reader checking the block number sequence and
     * throwing an exception is it's not sequential and starting with 1?
     * @return <code>true</code> if checking block number sequence, else <code>false</code>
     */
    public boolean checkBlockNumberSequence() {
        return checkBlockNumberSequence;
    }

    /**
     * Get the evio version number.
     * @return evio version number.
     */
    public int getEvioVersion() {
        return evioVersion;
    }

    /**
      * Get the path to the file.
      * @return path to the file
      */
     public String getPath() {
         return path;
     }

    /**
     * Get the file/buffer parser.
     * @return file/buffer parser.
     */
    public EventParser getParser() {
        return parser;
    }

    /**
     * Set the file/buffer parser.
     * @param parser file/buffer parser.
     */
    public void setParser(EventParser parser) {
        if (parser != null) {
            this.parser = parser;
        }
    }

     /**
     * Get the XML format dictionary is there is one.
     *
     * @return XML format dictionary, else null.
     */
    public String getDictionaryXML() {
        return dictionaryXML;
    }

    /**
     * Does this evio file have an associated XML dictionary?
     *
     * @return <code>true</code> if this evio file has an associated XML dictionary,
     *         else <code>false</code>
     */
    public boolean hasDictionaryXML() {
        return dictionaryXML != null;
    }

    /**
     * Get the number of events remaining in the file.
     *
     * @return number of events remaining in the file
     * @throws IOException if failed file access
     * @throws EvioException if failed reading from coda v3 file
     */
    public int getNumEventsRemaining() throws IOException, EvioException {
        return getEventCount() - eventNumber;
    }

	/**
	 * Maps the file into memory. The data are not actually loaded in memory-- subsequent reads will read
	 * random-access-like from the file.
	 *
     * @param inputChannel the input channel.
     * @throws IOException if file cannot be opened
	 */
	private synchronized void mapFile(FileChannel inputChannel) throws IOException {
		long sz = inputChannel.size();
		mappedByteBuffer = inputChannel.map(FileChannel.MapMode.READ_ONLY, 0L, sz);
        byteBuffer = mappedByteBuffer;
	}

    /**
     * Get the memory mapped buffer corresponding to the event file.
     * @return the memory mapped buffer corresponding to the event file.
     */
    public MappedByteBuffer getMappedByteBuffer() {
        return mappedByteBuffer;
    }

    /**
     * Get the byte buffer being read directly or corresponding to the event file.
     * @return the byte buffer being read directly or corresponding to the event file.
     */
    public ByteBuffer getByteBuffer() {
        return byteBuffer;
    }

    /**
     * Get the size of the file being read, in bytes.
     * For small files, obtain the file size using the memory mapped buffer's capacity.
     * @return the file size in bytes
     */
    public long fileSize() {
        return fileSize;
    }

    /**
     * Generate a table (ArrayList) of positions of events in file/buffer.
     * This method does <b>not</b> affect the byteBuffer position, eventNumber,
     * or lastBlock values. Valid only in versions 4 and later.
     *
     * @throws IOException if trouble reading file
     * @throws EvioException if version 3 or earlier
     */
    private void generateEventPositionTable() throws IOException, EvioException {

        if (evioVersion < 4) {
            throw new EvioException("Unsupported version (" + evioVersion + ")");
        }

        int      i, len, byteLen, blockEventCount, blockHdrSize;
        long     bytesLeft, position, firstPosition=0L;
        boolean  curLastBlock=false, firstBlock=true, hasDictionary=false;

//        long t2, t1 = System.currentTimeMillis();

        // Remove previously stored positions
        eventPositions.clear();

        if (sequentialRead) {
            // Start at the beginning of file, but record where we
            // were before scanning the file.
            firstPosition = fileChannel.position();
            if (firstPosition != 0L) {
                fileChannel.position(0L);
            }
            position  = 0L;
            bytesLeft = fileSize;
        }
        else {
            // Start at the beginning of byteBuffer without changing
            // its current position. Do this with absolute gets.
            position  = initialPosition;
            bytesLeft = byteBuffer.limit() - position;
        }

        // Keep track of the # of blocks & events in file/buffer
        blockCount = 0;
        eventCount = 0;

        while (!curLastBlock) {

            // File is now positioned before block header.
            // Look at block header to get info.
            if (sequentialRead) {
                fileChannel.position(position + 4*BlockHeaderV4.EV_VERSION);
                i               = dataStream.readInt();
                fileChannel.position(position + 4*BlockHeaderV4.EV_HEADERSIZE);
                blockHdrSize    = dataStream.readInt();
                fileChannel.position(position + 4*BlockHeaderV4.EV_COUNT);
                blockEventCount = dataStream.readInt();

                if (swap) {
                    i = Integer.reverseBytes(i);
                    blockHdrSize = Integer.reverseBytes(blockHdrSize);
                    blockEventCount = Integer.reverseBytes(blockEventCount);
                }
            }
            else {
                // Swapping is taken care of
                i               = byteBuffer.getInt((int)position + 4*BlockHeaderV4.EV_VERSION);
                blockHdrSize    = byteBuffer.getInt((int)position + 4*BlockHeaderV4.EV_HEADERSIZE);
                blockEventCount = byteBuffer.getInt((int)position + 4*BlockHeaderV4.EV_COUNT);
            }

            blockCount++;
            eventCount  += blockEventCount;
            curLastBlock = BlockHeaderV4.isLastBlock(i);
            if (firstBlock) hasDictionary = BlockHeaderV4.hasDictionary(i);

            // Hop over block header to data
            position  += 4*blockHdrSize;
            bytesLeft -= 4*blockHdrSize;
            if (sequentialRead) fileChannel.position(position);

//System.out.println("    hopped blk hdr, bytesLeft = " + bytesLeft + ", pos = " + position);

            // Check for a dictionary - the first event in the first block.
            // It's not included in the header block count, but we must take
            // it into account by skipping over it.
            if (firstBlock && hasDictionary) {
                firstBlock = false;

                // Get its length - bank's len does not include itself
                if (sequentialRead) {
                    len = dataStream.readInt();
                    if (swap) len = Integer.reverseBytes(len);
                    byteLen = 4*(len + 1);
                }
                else {
                    byteLen = 4*(byteBuffer.getInt((int)position) + 1);
                }

                // Skip over dictionary
                position  += byteLen;
                bytesLeft -= byteLen;
//System.out.println("    hopped dict, bytesLeft = " + bytesLeft + ", pos = " + position);
                if (sequentialRead) fileChannel.position(position);
            }

            // For each event in block, store its location
            for (i=0; i < blockEventCount; i++) {
                // Sanity check - must have at least 1 header's amount left
                if (bytesLeft < 8) {
                    throw new EvioException("File/buffer bad format");
                }

                // Store current position
                eventPositions.add(position);

                // Get length of current event (including full header)
                if (sequentialRead) {
                    len = dataStream.readInt();
                    if (swap) len = Integer.reverseBytes(len);
                    byteLen = 4*(len + 1);
                }
                else {
                    byteLen = 4*(byteBuffer.getInt((int)position) + 1);
                }

                position  += byteLen;
                bytesLeft -= byteLen;
//System.out.println("    hopped event, bytesLeft = " + bytesLeft + ", pos = " + position + "\n");
                if (sequentialRead) fileChannel.position(position);
            }
        }

        // Set the file position back to where it was when we started
        if (sequentialRead) fileChannel.position(firstPosition);

//        t2 = System.currentTimeMillis();
//        System.out.println("Time to scan file = " + (t2-t1) + " milliseconds");
    }


    /**
     * Reads part of the first block (physical record) header in order to determine
     * the evio version # and endianness of the file or buffer in question. These things
     * do <b>not</b> need to be examined in subsequent block headers.
     *
     * @return status of read attempt
     * @throws IOException if file access problems
     */
    protected synchronized ReadStatus getFirstHeader() throws IOException {

        // Have enough remaining bytes to read header?
        if (sequentialRead) {
            fileChannel.position(initialPosition);
            if (fileSize - fileChannel.position() < 32L) {
                return ReadStatus.END_OF_FILE;
            }
        }
        else {
            byteBuffer.position(initialPosition);
            if (byteBuffer.remaining() < 32) {
                byteBuffer.clear();
                return ReadStatus.END_OF_FILE;
            }
        }

        try {
            // Set the byte order to match the file's ordering.

            // Check the magic number for endianness. This requires
            // peeking ahead 7 ints or 28 bytes. Set the endianness
            // once we figure out what it is (buffer defaults to big endian).
            if (sequentialRead) {
                byte[] header = new byte[32]; // 4*8 = 32
                dataStream.readFully(header);
                // Borrow the byteBuffer variable since it isn't
                // used for anything else if doing sequential read.
                byteBuffer = ByteBuffer.wrap(header);
                byteOrder = ByteOrder.BIG_ENDIAN;
            }
            else {
                byteOrder = byteBuffer.order();
            }

            int magicNumber = byteBuffer.getInt(MAGIC_OFFSET);

            if (magicNumber != IBlockHeader.MAGIC_NUMBER) {
                swap = true;

                if (byteOrder == ByteOrder.BIG_ENDIAN) {
                    byteOrder = ByteOrder.LITTLE_ENDIAN;
                }
                else {
                    byteOrder = ByteOrder.BIG_ENDIAN;
                }
                byteBuffer.order(byteOrder);

                // Reread magic number to make sure things are OK
                magicNumber = byteBuffer.getInt(MAGIC_OFFSET);
                if (magicNumber != IBlockHeader.MAGIC_NUMBER) {
System.out.println("ERROR reread magic # (" + magicNumber + ") & still not right");
                    return ReadStatus.EVIO_EXCEPTION;
                }
            }

            // Check the version number. This requires peeking ahead 5 ints or 20 bytes.
            evioVersion = byteBuffer.getInt(VERSION_OFFSET) & VERSION_MASK;
            if (evioVersion < 1)  {
                return ReadStatus.EVIO_EXCEPTION;
            }
//System.out.println("Evio version# = " + evioVersion);

            if (sequentialRead) {
                // Create a buffer to hold the entire first block of data
                int blkSize = byteBuffer.getInt(BLOCK_SIZE_OFFSET);
                if (blockBuffer != null && blockBuffer.capacity() >= 4*blkSize) {
                    blockBuffer.clear();
                    blockBuffer.limit(4*blkSize);
//System.out.println("Use existing ByteBuffer of size = " + blockBuffer.capacity());
                }
                else {
                    // Make this bigger than necessary so we're not constantly reallocating
                    blockBuffer = ByteBuffer.allocate(4*blkSize + 10000);
                    blockBuffer.limit(4*blkSize);
                    blockBuffer.order(byteOrder);
//System.out.println("create new ByteBuffer of size = " + 4*blkSize + " bytes");
                }

                // Read the entire first block of data
                // First copy over the header we just read in.
                // Then read rest of the block.
                blockBuffer.put(byteBuffer);
                fileChannel.read(blockBuffer);
                // Get it ready for reading
                blockBuffer.flip();
                // Convenience variable
                byteBuffer = blockBuffer;
            }

            try {
                if (evioVersion >= 4) {
                    // Cache the starting position
                    // Here we use position in blockBuffer if sequential
                    blockHeader4.setBufferStartingPosition(byteBuffer.position());

//                    int pos = byteBuffer.position();
//System.out.println("BlockHeader v4:");
//System.out.println("   block length  = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=4;
//System.out.println("   block number  = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=4;
//System.out.println("   header length = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=4;
//System.out.println("   event count   = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=8;
//System.out.println("   version       = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=8;
//System.out.println("   magic number  = 0x" + Integer.toHexString(byteBuffer.getInt(pos))); pos+=4;
//System.out.println();
//
//
                    // Read the header data.
                    blockHeader4.setSize(byteBuffer.getInt());
                    blockHeader4.setNumber(byteBuffer.getInt());
                    blockHeader4.setHeaderLength(byteBuffer.getInt());
                    blockHeader4.setEventCount(byteBuffer.getInt());
                    blockHeader4.setReserved1(byteBuffer.getInt());

                    // Use 6th word to set bit info & version
                    blockHeader4.parseToBitInfo(byteBuffer.getInt());
                    blockHeader4.setVersion(evioVersion);
                    lastBlock = blockHeader4.getBitInfo(1);
                    blockHeader4.setReserved2(byteBuffer.getInt());
                    blockHeader4.setMagicNumber(byteBuffer.getInt());
                    blockHeader = blockHeader4;

                    // Deal with non-standard header lengths here
                    int headerLenDiff = blockHeader4.getHeaderLength() - BlockHeaderV4.HEADER_SIZE;
                    // If too small quit with error since headers have a minimum size
                    if (headerLenDiff < 0) {
                        return ReadStatus.EVIO_EXCEPTION;
                    }
                    // If bigger, read extra ints & discard since we don't know what else to do
                    else if (headerLenDiff > 0) {
                        for (int i=0; i < headerLenDiff; i++) {
//System.out.println("Getting extra header int");
                            byteBuffer.getInt();
                        }
                    }

//System.out.println("BlockHeader v4:");
//System.out.println("   block length  = " + blockHeader4.getSize() + " ints");
//System.out.println("   block number  = " + blockHeader4.getNumber());
//System.out.println("   header length = " + blockHeader4.getHeaderLength() + " ints");
//System.out.println("   event count   = " + blockHeader4.getEventCount());
//System.out.println("   version       = " + blockHeader4.getVersion());
//System.out.println("   has Dict      = " + blockHeader4.getBitInfo(0));
//System.out.println("   is End        = " + lastBlock);
//System.out.println("   magic number  = " + Integer.toHexString(blockHeader4.getMagicNumber()));
//System.out.println();

                    // Is there a dictionary? If so, read it now.
                    if (blockHeader4.hasDictionary()) {
                        readDictionary();
                    }
                }
                else {
                    // cache the starting position
                    blockHeader2.setBufferStartingPosition(byteBuffer.position());

                    // read the header data.
                    blockHeader2.setSize(byteBuffer.getInt());
                    blockHeader2.setNumber(byteBuffer.getInt());
                    blockHeader2.setHeaderLength(byteBuffer.getInt());
                    blockHeader2.setStart(byteBuffer.getInt());
                    blockHeader2.setEnd(byteBuffer.getInt());
                    // skip version
                    byteBuffer.getInt();
                    blockHeader2.setVersion(evioVersion);
                    blockHeader2.setReserved1(byteBuffer.getInt());
                    blockHeader2.setMagicNumber(byteBuffer.getInt());
                    blockHeader = blockHeader2;
                }

                // Store this for later regurgitation of blockCount
                firstBlockSize = 4*blockHeader.getSize();

                // check block number if so configured
                if (checkBlockNumberSequence) {
                    if (blockHeader.getNumber() != blockNumberExpected) {

System.out.println("block # out of sequence, got " + blockHeader.getNumber() +
                           " expecting " + blockNumberExpected);

                        return ReadStatus.EVIO_EXCEPTION;
                    }
                    blockNumberExpected++;
                }
            }
            catch (EvioException e) {
                e.printStackTrace();
                return ReadStatus.EVIO_EXCEPTION;
            }

        }
        catch (BufferUnderflowException a) {
System.err.println("ERROR endOfBuffer " + a);
            byteBuffer.clear();
            return ReadStatus.UNKNOWN_ERROR;
        }

        if (!sequentialRead) {
            byteBuffer.position(initialPosition);
        }

        return ReadStatus.SUCCESS;
    }


    /**
     * Reads the block (physical record) header. Assumes the mapped buffer or file is positioned
     * at the start of the next block header (physical record.) By the time this is called,
     * the version # and byte order have already been determined. Not necessary to do that
     * for each block header that's read.<br>
     *
     * A Bank header is 8, 32-bit ints. The first int is the size of the block in ints
     * (not counting the length itself, i.e., the number of ints to follow).
     *
     * Most users should have no need for this method, since most applications do not
     * care about the block (physical record) header.
     *
     * @return status of read attempt
     * @throws IOException if file access problems
     */
    protected synchronized ReadStatus nextBlockHeader() throws IOException {

        // We already read the last block header
        if (lastBlock) {
            return ReadStatus.END_OF_FILE;
        }

        // Have enough remaining bytes to read header?
        if (sequentialRead) {
            if (fileSize - fileChannel.position() < 32L) {
                return ReadStatus.END_OF_FILE;
            }
        }
        else {
            if (byteBuffer.remaining() < 32) {
                byteBuffer.clear();
                return ReadStatus.END_OF_FILE;
            }
        }
        try {

            if (sequentialRead) {
                // Read len of block
                int blkSize = dataStream.readInt();
                if (swap) blkSize = Integer.reverseBytes(blkSize);
                // Create a buffer to hold the entire first block of data
                if (blockBuffer != null && blockBuffer.capacity() >= 4*blkSize) {
                    blockBuffer.clear();
                    blockBuffer.limit(4*blkSize);
                }
                else {
                    // Make this bigger than necessary so we're not constantly reallocating
                    blockBuffer = ByteBuffer.allocate(4*blkSize + 10000);
                    blockBuffer.limit(4*blkSize);
                    blockBuffer.order(byteOrder);
                }
                // Read the entire first block of data
                blockBuffer.putInt(blkSize);
                fileChannel.read(blockBuffer);
                blockBuffer.flip();
                // Convenience variable
                byteBuffer = blockBuffer;
                // Now keeping track of pos in this new blockBuffer (should be 0)
                blockHeader.setBufferStartingPosition(blockBuffer.position());
            }
            else {
                // Record starting position
                blockHeader.setBufferStartingPosition(byteBuffer.position());
            }

            if (evioVersion >= 4) {
                // Read the header data.
                blockHeader4.setSize(byteBuffer.getInt());
                blockHeader4.setNumber(byteBuffer.getInt());
                blockHeader4.setHeaderLength(byteBuffer.getInt());
                blockHeader4.setEventCount(byteBuffer.getInt());
                blockHeader4.setReserved1(byteBuffer.getInt());
                // Use 6th word to set bit info
                blockHeader4.parseToBitInfo(byteBuffer.getInt());
                blockHeader4.setVersion(evioVersion);
                lastBlock = blockHeader4.getBitInfo(1);
                blockHeader4.setReserved2(byteBuffer.getInt());
                blockHeader4.setMagicNumber(byteBuffer.getInt());
                blockHeader = blockHeader4;

                // Deal with non-standard header lengths here
                int headerLenDiff = blockHeader4.getHeaderLength() - BlockHeaderV4.HEADER_SIZE;
                // If too small quit with error since headers have a minimum size
                if (headerLenDiff < 0) {
                    return ReadStatus.EVIO_EXCEPTION;
                }
                // If bigger, read extra ints
                else if (headerLenDiff > 0) {
                    for (int i=0; i < headerLenDiff; i++) {
                        byteBuffer.getInt();
                    }
                }

//System.out.println("BlockHeader v4:");
//System.out.println("   block length  = " + blockHeader4.getSize() + " ints");
//System.out.println("   block number  = " + blockHeader4.getNumber());
//System.out.println("   header length = " + blockHeader4.getHeaderLength() + " ints");
//System.out.println("   event count   = " + blockHeader4.getEventCount());
//System.out.println("   version       = " + blockHeader4.getVersion());
//System.out.println("   has Dict      = " + blockHeader4.getBitInfo(0));
//System.out.println("   is End        = " + lastBlock);
//System.out.println("   magic number  = " + Integer.toHexString(blockHeader4.getMagicNumber()));
//System.out.println();

            }
            else if (evioVersion < 4) {
                // read the header data.
                blockHeader2.setSize(byteBuffer.getInt());
                blockHeader2.setNumber(byteBuffer.getInt());
                blockHeader2.setHeaderLength(byteBuffer.getInt());
                blockHeader2.setStart(byteBuffer.getInt());
                blockHeader2.setEnd(byteBuffer.getInt());
                // skip version
                byteBuffer.getInt();
                blockHeader2.setVersion(evioVersion);
                blockHeader2.setReserved1(byteBuffer.getInt());
                blockHeader2.setMagicNumber(byteBuffer.getInt());
                blockHeader = blockHeader2;
            }
            else {
                // bad version # - should never happen
                return ReadStatus.EVIO_EXCEPTION;
            }

            // check block number if so configured
            if (checkBlockNumberSequence) {
                if (blockHeader.getNumber() != blockNumberExpected) {

System.out.println("block # out of sequence, got " + blockHeader.getNumber() +
                   " expecting " + blockNumberExpected);

                    return ReadStatus.EVIO_EXCEPTION;
                }
                blockNumberExpected++;
            }
        }
        catch (EvioException e) {
            e.printStackTrace();
            return ReadStatus.EVIO_EXCEPTION;
        }
        catch (BufferUnderflowException a) {
System.err.println("ERROR endOfBuffer " + a);
            byteBuffer.clear();
            return ReadStatus.UNKNOWN_ERROR;
        }

        return ReadStatus.SUCCESS;
    }


    /**
     * This method is only called once at the very beginning if buffer is known to have
     * a dictionary. It then reads that dictionary. Only called in versions 4 & up.
     *
     * @return <code>true</code> if dictionary was found & read, else <code>false</code>.
     * @throws java.io.IOException if file access problems
     * @throws org.jlab.coda.jevio.EvioException if failed read due to bad buffer format;
     *                       if version 3 or earlier
     */
     private synchronized void readDictionary() throws IOException, EvioException {

         if (evioVersion < 4) {
             throw new EvioException("Unsupported version (" + evioVersion + ")");
         }

         // How many bytes remain in this block?
         int bytesRemaining = blockBytesRemaining();
         if (bytesRemaining < 12) {
             throw new EvioException("Not enough data in first block");
         }

         // Once here, we are assured the entire next event is in this block.
         int length;
         length = byteBuffer.getInt();

         if (length < 1) {
             throw new EvioException("Bad value for dictionary length (too big for java?)");
         }
         bytesRemaining -= 4;

         // If we've already read in the dictionary,
         // don't bother doing it again just hop over it.
         if (dictionaryXML != null) {
            byteBuffer.position(byteBuffer.position() + 4 + 4 * (length - 1));
             return;
         }

         // Since we're only interested in length, read but ignore rest of the header.
         byteBuffer.getInt();
         bytesRemaining -= 4;

         // get the raw data
         int eventDataSizeBytes = 4*(length - 1);
         if (bytesRemaining < eventDataSizeBytes) {
             throw new EvioException("Not enough data in first block");
         }

         byte bytes[] = new byte[eventDataSizeBytes];

         // Read in dictionary data
         try {
            byteBuffer.get(bytes, 0, eventDataSizeBytes);
         }
         catch (Exception e) {
             throw new EvioException("Problems reading buffer");
         }

         // This is the very first event and must be a dictionary
         String[] strs = BaseStructure.unpackRawBytesToStrings(bytes, 0);
         if (strs == null) {
             throw new EvioException("Data in bad format");
         }
         dictionaryXML = strs[0];
     }

    /**
     * Get the event in the file/buffer at a given index (starting at 1).
     * As useful as this sounds, most applications will probably call
     * {@link #parseEvent(int)} instead, since it combines combines
     * getting the event with parsing it.<p>
     *
     * @param  index number of event desired, starting at 1, from beginning of file/buffer
     * @return the event in the file at the given index or null if none
     * @throws java.io.IOException if failed file access
     * @throws org.jlab.coda.jevio.EvioException if failed read due to bad file/buffer format;
     *                       if out of memory; if index < 1
     */
    public synchronized EvioEvent getEvent(int index) throws IOException, EvioException {

        if (index < 1) {
            throw new EvioException("index arg starts at 1");
        }

        if (index > eventPositions.size()) {
            return null;
        }

        if (evioVersion < 4) {
            return gotoEventNumber(index);
        }

        index--;

        EvioEvent event = new EvioEvent();
        BaseStructureHeader header = event.getHeader();

        long origPosition;
        int length, eventDataSizeBytes = 0;

        if (sequentialRead) {
            // Save the current position
            origPosition = fileChannel.position();

            // Go to the correct position for the desired event
            long position = eventPositions.get(index);
            fileChannel.position(position);

            // Read length of event
            length = dataStream.readInt();
            if (swap) length = Integer.reverseBytes(length);
        }
        else {
            origPosition = byteBuffer.position();
            long position = eventPositions.get(index);
            byteBuffer.position((int)position);
            length = byteBuffer.getInt();
        }

        if (length < 1) {
            throw new EvioException("Bad file/buffer format");
        }
        header.setLength(length);

        try {
            // Read second header word
            if (byteOrder == ByteOrder.BIG_ENDIAN) {
                // Interested in bit pattern, not negative numbers
                int dt;
                if (sequentialRead) {
                    short s = dataStream.readShort();
                    if (swap) s = Short.reverseBytes(s);
                    header.setTag(ByteDataTransformer.shortBitsToInt(s));
                    dt = ByteDataTransformer.byteBitsToInt(dataStream.readByte());
                }
                else {
                    header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
                    dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());
                }

                int type = dt & 0x3f;
                int padding = dt >>> 6;
                // If only 7th bit set, that can only be the legacy tagsegment type
                // with no padding information - convert it properly.
                if (dt == 0x40) {
                    type = DataType.TAGSEGMENT.getValue();
                    padding = 0;
                }
                header.setDataType(type);
                header.setPadding(padding);

                // Once we know what the data type is, let the no-arg constructed
                // event know what type it is holding so xml names are set correctly.
                event.setXmlNames();
                if (sequentialRead) {
                    header.setNumber(ByteDataTransformer.byteBitsToInt(dataStream.readByte()));
                }
                else {
                    header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
                }
            }
            else {
                int dt;
                if (sequentialRead) {
                    header.setNumber(ByteDataTransformer.byteBitsToInt(dataStream.readByte()));
                    dt = ByteDataTransformer.byteBitsToInt(dataStream.readByte());
                }
                else {
                    header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
                    dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());
                }

                int type = dt & 0x3f;
                int padding = dt >>> 6;
                if (dt == 0x40) {
                    type = DataType.TAGSEGMENT.getValue();
                    padding = 0;
                }
                header.setDataType(type);
                header.setPadding(padding);

                event.setXmlNames();
                if (sequentialRead) {
                    short s = dataStream.readShort();
                    if (swap) s = Short.reverseBytes(s);
                    header.setTag(ByteDataTransformer.shortBitsToInt(s));
                }
                else {
                    header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
                }
            }

            // Read the raw data
            eventDataSizeBytes = 4*(length - 1);
            byte bytes[] = new byte[eventDataSizeBytes];
            if (sequentialRead) {
                dataStream.readFully(bytes, 0, eventDataSizeBytes);
            }
            else {
                byteBuffer.get(bytes, 0, eventDataSizeBytes);
            }

            event.setRawBytes(bytes);
            event.setByteOrder(byteOrder);
            event.setEventNumber(index + 1);

        }
        catch (OutOfMemoryError e) {
            throw new EvioException("Out Of Memory: (event size = " + eventDataSizeBytes + ")", e);
        }
        catch (Exception e) {
            throw new EvioException("Error", e);
        }
        finally {
            // Reset the file/buffer position in all cases
            if (sequentialRead) {
                fileChannel.position(origPosition);
            }
            else {
                byteBuffer.position((int)origPosition);
            }
        }

        return event;
    }


	/**
	 * This is the workhorse method. It retrieves the desired event from the file/buffer,
     * and then parses it SAX-like. It will drill down and uncover all structures
     * (banks, segments, and tagsegments) and notify any interested listeners.
	 *
     * @param  index number of event desired, starting at 1, from beginning of file/buffer
	 * @return the parsed event at the given index or null if none
     * @throws java.io.IOException if failed file access
     * @throws org.jlab.coda.jevio.EvioException if failed read due to bad file/buffer format;
     *                       if out of memory; if index < 1
	 */
	public synchronized EvioEvent parseEvent(int index) throws IOException, EvioException {
		EvioEvent event = getEvent(index);
        if (event != null) parseEvent(event);
		return event;
	}


    /**
     * Get the next event in the file. As useful as this sounds, most applications will probably call
     * {@link #parseNextEvent()} instead, since it combines combines getting the next
     * event with parsing the next event.<p>
     * In evio version 4, events no longer cross block boundaries. There are only one or more
     * complete events in each block. No changes were made to this method from versions 1-3 in order
     * to read the version 4 format as it is subset of versions 1-3 with variable block length.
     *
     * @return the next event in the file. On error it throws an EvioException.
     *         On end of file, it returns <code>null</code>.
     * @throws java.io.IOException if failed file access
     * @throws org.jlab.coda.jevio.EvioException if failed read due to bad buffer format
     */
    public synchronized EvioEvent nextEvent() throws IOException, EvioException {

        // Are we at the top of the file?
        boolean atTop = false;
        if (sequentialRead) {
            if (fileChannel.position() == 0L) {
                atTop = true;
            }
        }
        else {
            return getEvent(sequentialCounter++);
        }

        boolean isFirstEvent = false;
        if (atTop) {
            ReadStatus status = nextBlockHeader();
            if (status != ReadStatus.SUCCESS) {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
            isFirstEvent = true;
        }

        EvioEvent event = new EvioEvent();
        BaseStructureHeader header = event.getHeader();
        long currentPosition = byteBuffer.position();

        // How many bytes remain in this block until we reach the next block header?
        // Must see if we have to deal with crossing physical record boundaries (< version 4).
        int bytesRemaining = blockBytesRemaining();
        if (bytesRemaining < 0) {
            throw new EvioException("Number of block bytes remaining is negative.");
        }

        // Are we exactly at the end of the block (physical record)?
        if (bytesRemaining == 0) {
            ReadStatus status = nextBlockHeader();
            if (status == ReadStatus.SUCCESS) {
                return nextEvent();
            }
            else if (status == ReadStatus.END_OF_FILE) {
                return null;
            }
            else {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
        }
        // Or have we already read in the last event?
        // If jevio versions 1-3, the last block may not be full.
        // Thus bytesRemaining may be > 0, but we may have read
        // in all the existing data. (This should never happen in version 4).
        else if (blockHeader.getBufferEndingPosition() == currentPosition) {
            return null;
        }

        // Version   4: once here, we are assured the entire next event is in this block.
        // Version 1-3: no matter what, we can get the length of the next event.
        // A non positive length indicates eof;
        int length;
        length = byteBuffer.getInt();
        if (length < 1) {
            return null;
        }

        header.setLength(length);
        bytesRemaining -= 4; // just read in 4 bytes

        // Versions 1-3: if we were unlucky, after reading the length
        //               there are no bytes remaining in this bank.
        // Don't really need the "if (version < 4)" here except for clarity.
        if (evioVersion < 4) {
            if (bytesRemaining == 0) {
                ReadStatus status = nextBlockHeader();
                if (status == ReadStatus.END_OF_FILE) {
                    return null;
                }
                else if (status != ReadStatus.SUCCESS) {
                    throw new EvioException("Failed reading block header in nextEvent.");
                }
                bytesRemaining = blockBytesRemaining();
            }
        }

        // Now should be good to go, except data may cross block boundary.
        // In any case, should be able to read the rest of the header.
        if (byteOrder == ByteOrder.BIG_ENDIAN) {
            // interested in bit pattern, not negative numbers
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
            int dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());

            int type = dt & 0x3f;
            int padding = dt >>> 6;
            // If only 7th bit set, that can only be the legacy tagsegment type
            // with no padding information - convert it properly.
            if (dt == 0x40) {
                type = DataType.TAGSEGMENT.getValue();
                padding = 0;
            }
            header.setDataType(type);
            header.setPadding(padding);

            // Once we know what the data type is, let the no-arg constructed
            // event know what type it is holding so xml names are set correctly.
            event.setXmlNames();
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
        }
        else {
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            int dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());

            int type = dt & 0x3f;
            int padding = dt >>> 6;
            if (dt == 0x40) {
                type = DataType.TAGSEGMENT.getValue();
                padding = 0;
            }
            header.setDataType(type);
            header.setPadding(padding);

            event.setXmlNames();
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
        }
        bytesRemaining -= 4; // just read in 4 bytes

        // get the raw data
        int eventDataSizeBytes = 4*(length - 1);

        try {
            byte bytes[] = new byte[eventDataSizeBytes];

            int bytesToGo = eventDataSizeBytes;
            int offset = 0;

            // Don't really need the "if (version < 4)" here except for clarity.
            if (evioVersion < 4) {
                // be in while loop if have to cross block boundary[ies].
                while (bytesToGo > bytesRemaining) {
                    byteBuffer.get(bytes, offset, bytesRemaining);

                    ReadStatus status = nextBlockHeader();
                    if (status == ReadStatus.END_OF_FILE) {
                        return null;
                    }
                    else if (status != ReadStatus.SUCCESS) {
                        throw new EvioException("Failed reading block header after crossing boundary in nextEvent.");
                    }
                    bytesToGo -= bytesRemaining;
                    offset += bytesRemaining;
                    bytesRemaining = blockBytesRemaining();
                }
            }

            // last (perhaps only) read
            byteBuffer.get(bytes, offset, bytesToGo);
            event.setRawBytes(bytes);
            event.setByteOrder(byteOrder); // add this to track endianness, timmer

            // if this is the very first event, check to see if it's a dictionary
            if (isFirstEvent && blockHeader.hasDictionary()) {
                // if we've NOT already been through this before, extract the dictionary
                if (dictionaryXML == null) {
                    dictionaryXML = event.getStringData()[0];
                }
                // give the user the next event, not the dictionary
                return nextEvent();
            }
            else {
                event.setEventNumber(++eventNumber);
                return event;
            }
        }
        catch (OutOfMemoryError ome) {
            System.err.println("Out Of Memory\n" +
                                       "eventDataSizeBytes = " + eventDataSizeBytes + "\n" +
                                       "bytes Remaining = " + bytesRemaining + "\n" +
                                       "event Count: " + eventCount);
            return null;
        }
        catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Get the next event in the file. As useful as this sounds, most applications will probably call
     * {@link #parseNextEvent()} instead, since it combines combines getting the next
     * event with parsing the next event.<p>
     * In evio version 4, events no longer cross block boundaries. There are only one or more
     * complete events in each block. No changes were made to this method from versions 1-3 in order
     * to read the version 4 format as it is subset of versions 1-3 with variable block length.
     *
     * @return the next event in the file. On error it throws an EvioException.
     *         On end of file, it returns <code>null</code>.
     * @throws java.io.IOException if failed file access
     * @throws org.jlab.coda.jevio.EvioException if failed read due to bad buffer format
     */
    public synchronized EvioEvent nextEventOrig() throws IOException, EvioException {
        EvioEvent event = new EvioEvent();
        BaseStructureHeader header = event.getHeader();

        long currentPosition = byteBuffer.position();

        // Are we at the top of the file?
        boolean atTop = false;
        boolean isFirstEvent = false;
        if (sequentialRead) {
            if (fileChannel.position() == 0L) {
                atTop = true;
            }
        }
        else if (byteBuffer.position() == 0) {
            atTop = true;
        }

        if (atTop) {
            ReadStatus status = nextBlockHeader();
            if (status != ReadStatus.SUCCESS) {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
            isFirstEvent = true;
        }

        // How many bytes remain in this block until we reach the next block header?
        // Must see if we have to deal with crossing physical record boundaries (< version 4).
        int bytesRemaining = blockBytesRemaining();
        if (bytesRemaining < 0) {
            throw new EvioException("Number of block bytes remaining is negative.");
        }

        // Are we exactly at the end of the block (physical record)?
        if (bytesRemaining == 0) {
            ReadStatus status = nextBlockHeader();
            if (status == ReadStatus.SUCCESS) {
                return nextEvent();
            }
            else if (status == ReadStatus.END_OF_FILE) {
                return null;
            }
            else {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
        }
        // Or have we already read in the last event?
        // If jevio versions 1-3, the last block may not be full.
        // Thus bytesRemaining may be > 0, but we may have read
        // in all the existing data. (This should never happen in version 4).
        else if (blockHeader.getBufferEndingPosition() == currentPosition) {
            return null;
        }

        // Version   4: once here, we are assured the entire next event is in this block.
        // Version 1-3: no matter what, we can get the length of the next event.
        // A non positive length indicates eof;
        int length;
        length = byteBuffer.getInt();
        if (length < 1) {
            return null;
        }

        header.setLength(length);
        bytesRemaining -= 4; // just read in 4 bytes

        // Versions 1-3: if we were unlucky, after reading the length
        //               there are no bytes remaining in this bank.
        // Don't really need the "if (version < 4)" here except for clarity.
        if (evioVersion < 4) {
            if (bytesRemaining == 0) {
                ReadStatus status = nextBlockHeader();
                if (status == ReadStatus.END_OF_FILE) {
                    return null;
                }
                else if (status != ReadStatus.SUCCESS) {
                    throw new EvioException("Failed reading block header in nextEvent.");
                }
                bytesRemaining = blockBytesRemaining();
            }
        }

        // Now should be good to go, except data may cross block boundary.
        // In any case, should be able to read the rest of the header.
        if (byteOrder == ByteOrder.BIG_ENDIAN) {
            // interested in bit pattern, not negative numbers
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
            int dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());

            int type = dt & 0x3f;
            int padding = dt >>> 6;
            // If only 7th bit set, that can only be the legacy tagsegment type
            // with no padding information - convert it properly.
            if (dt == 0x40) {
                type = DataType.TAGSEGMENT.getValue();
                padding = 0;
            }
            header.setDataType(type);
            header.setPadding(padding);

            // Once we know what the data type is, let the no-arg constructed
            // event know what type it is holding so xml names are set correctly.
            event.setXmlNames();
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
        }
        else {
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            int dt = ByteDataTransformer.byteBitsToInt(byteBuffer.get());

            int type = dt & 0x3f;
            int padding = dt >>> 6;
            if (dt == 0x40) {
                type = DataType.TAGSEGMENT.getValue();
                padding = 0;
            }
            header.setDataType(type);
            header.setPadding(padding);

            event.setXmlNames();
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
        }
        bytesRemaining -= 4; // just read in 4 bytes

        // get the raw data
        int eventDataSizeBytes = 4*(length - 1);

        try {
            byte bytes[] = new byte[eventDataSizeBytes];

            int bytesToGo = eventDataSizeBytes;
            int offset = 0;

            // Don't really need the "if (version < 4)" here except for clarity.
            if (evioVersion < 4) {
                // be in while loop if have to cross block boundary[ies].
                while (bytesToGo > bytesRemaining) {
                    byteBuffer.get(bytes, offset, bytesRemaining);

                    ReadStatus status = nextBlockHeader();
                    if (status == ReadStatus.END_OF_FILE) {
                        return null;
                    }
                    else if (status != ReadStatus.SUCCESS) {
                        throw new EvioException("Failed reading block header after crossing boundary in nextEvent.");
                    }
                    bytesToGo -= bytesRemaining;
                    offset += bytesRemaining;
                    bytesRemaining = blockBytesRemaining();
                }
            }

            // last (perhaps only) read
            byteBuffer.get(bytes, offset, bytesToGo);
            event.setRawBytes(bytes);
            event.setByteOrder(byteOrder); // add this to track endianness, timmer

            // if this is the very first event, check to see if it's a dictionary
            if (isFirstEvent && blockHeader.hasDictionary()) {
                // if we've NOT already been through this before, extract the dictionary
                if (dictionaryXML == null) {
                    dictionaryXML = event.getStringData()[0];
                }
                // give the user the next event, not the dictionary
                return nextEvent();
            }
            else {
                event.setEventNumber(++eventNumber);
                return event;
            }
        }
        catch (OutOfMemoryError ome) {
            System.err.println("Out Of Memory\n" +
                                       "eventDataSizeBytes = " + eventDataSizeBytes + "\n" +
                                       "bytes Remaining = " + bytesRemaining + "\n" +
                                       "event Count: " + eventCount);
            return null;
        }
        catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

	/**
	 * This is the workhorse method. It retrieves the next event from the file, and then parses is SAX-like. It will
	 * drill down and uncover all structures (banks, segments, and tagsegments) and notify any interested listeners.
	 *
	 * @return the event that was parsed. On error it throws an EvioException. On end of file, it returns
	 *         <code>null</code>.
     * @throws java.io.IOException if failed file access
     * @throws org.jlab.coda.jevio.EvioException if read failure or bad format
	 */
	public synchronized EvioEvent parseNextEvent() throws IOException, EvioException {
		EvioEvent event = nextEvent();
		if (event != null) {
			parseEvent(event);
		}
		return event;
	}

	/**
	 * This will parse an event, SAX-like. It will drill down and uncover all structures (banks, segments, and
	 * tagsegments) and notify any interested listeners.
	 *
	 * As useful as this sounds, most applications will probably call {@link #parseNextEvent()} instead,
	 * since it combines combines getting the next event with parsing the next event .
	 *
	 * @param evioEvent the event to parse.
	 * @throws org.jlab.coda.jevio.EvioException if bad format
	 */
	public synchronized void parseEvent(EvioEvent evioEvent) throws EvioException {
		parser.parseEvent(evioEvent);
	}

	/**
	 * Get the number of bytes remaining in the current block (physical record). This is used for pathology checks like
	 * crossing the block boundary.
	 *
	 * @return the number of bytes remaining in the current block (physical record).
     */
	private int blockBytesRemaining() {
		try {
            return blockHeader.bytesRemaining(byteBuffer.position());
		}
		catch (EvioException e) {
			e.printStackTrace();
			return -1;
		}
	}

	/**
	 * The equivalent of rewinding the file. What it actually does
     * is set the position of the file/buffer back to where it was
     * after calling the constructor - after the first block.
     * This method, along with the two <code>position()</code> and the
     * <code>close()</code> method, allows applications to treat files
     * in a normal random access manner.
     *
     * @throws java.io.IOException if failed file access or buffer/file read
	 */
    public void rewind() throws IOException {
        if (sequentialRead) {
            fileChannel.position(initialPosition);
        }
		else {
            byteBuffer.position(initialPosition);
        }

        lastBlock = false;
		eventNumber = 0;
        blockNumberExpected = 1;
        sequentialCounter   = 1;
//        blockHeader.setBufferStartingPosition(initialPosition);

        // Read first block header
        if (getFirstHeader() != ReadStatus.SUCCESS) {
            throw new IOException("Failed reading first block header/dictionary");
        }
	}

	/**
	 * This is equivalent to obtaining the current position in the file.
     * What it actually does is return the position of the buffer. This
     * method, along with the <code>rewind()</code>, <code>position(int)</code>
     * and the <code>close()</code> method, allows applications to treat files
     * in a normal random access manner.
	 *
	 * @return the position of the buffer.
     * @throws IOException if error accessing file
	 */
	public long position() throws IOException {
        if (sequentialRead) {
            return fileChannel.position();
        }
		return byteBuffer.position();
	}

	/**
	 * This method sets the current position in the file or buffer. This
     * method, along with the <code>rewind()</code>, <code>position()</code>
     * and the <code>close()</code> method, allows applications to treat files
     * in a normal random access manner.<p>
     *
     * <b>HOWEVER</b>, using this method is not necessary for random access of
     * events and is no longer recommended because it interferes with the sequential
     * reading of events. Therefore it is now deprecated.
     *
	 * @deprecated
	 * @param position the new position of the buffer.
     * @throws IOException if error accessing file
	 */
	public void position(long position) throws IOException  {
         if (sequentialRead) {
            fileChannel.position(position);
        }
        else {
            byteBuffer.position((int)position);
        }
	}

	/**
	 * This is closes the file, but for buffers it only sets the position to 0.
     * This method, along with the <code>rewind()</code> and the two
     * <code>position()</code> methods, allows applications to treat files
     * in a normal random access manner.
     *
     * @throws IOException if error accessing file
	 */
    public void close() throws IOException {
        if (sequentialRead) {
            fileChannel.close();
            dataStream.close();
        }
        else {
            byteBuffer.position(initialPosition);
        }
	}

	/**
	 * This returns the current (active) block (physical record) header.
     * Since most users have no interest in physical records, this method
     * should not be used often. Mostly it is used by the test programs in the
	 * <code>EvioReaderTest</code> class.
	 *
	 * @return the current block header.
	 */
	public IBlockHeader getCurrentBlockHeader() {
		return blockHeader;
	}

	/**
	 * Go to a specific event in the file. The events are numbered 1..N.
     * This number is transient--it is not part of the event as stored in the evio file.
     * In versions 4 and up this is just a wrapper on {@link #getEvent(int)}.
	 *
	 * @param evNumber the event number in a 1..N counting sense, from the start of the file.
     * @return the specified event in file or null if there's an error or nothing at that event #.
     * @throws IOException if failed file access
	 */
	public EvioEvent gotoEventNumber(int evNumber) throws IOException {

		if (evNumber < 1) {
			return null;
		}

        if (evioVersion > 3) {
            try {
                return parseEvent(evNumber);
            }
            catch (EvioException e) {
                return null;
            }
        }

		rewind();
		EvioEvent event;

        int currentCount = 0;

		try {
			// get the first evNumber - 1 events without parsing
			for (int i = currentCount+1; i < evNumber; i++) {
				event = nextEvent();
				if (event == null) {
					throw new EvioException("Asked to go to event: " + evNumber + ", which is beyond the end of file");
				}
			}
			// get one more event, the evNumber'th event, with parsing
			return parseNextEvent();
		}
		catch (EvioException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
     * Rewrite the file to XML (not including dictionary).
	 *
	 * @param path the path to the XML file.
	 * @return the status of the write.
     * @throws IOException if failed file access
	 */
	public WriteStatus toXMLFile(String path) throws IOException {
		return toXMLFile(path, null);
	}

	/**
	 * Rewrite the file to XML (not including dictionary).
	 *
	 * @param path the path to the XML file.
	 * @param progressListener and optional progress listener, can be <code>null</code>.
	 * @return the status of the write.
     * @throws IOException if failed file access
	 * @see IEvioProgressListener
	 */
	public WriteStatus toXMLFile(String path, IEvioProgressListener progressListener)
                throws IOException {
		FileOutputStream fos;

		try {
			fos = new FileOutputStream(path);
		}
		catch (FileNotFoundException e) {
			e.printStackTrace();
			return WriteStatus.CANNOT_OPEN_FILE;
		}

		try {
			XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(fos);
			xmlWriter.writeStartDocument();
			xmlWriter.writeCharacters("\n");
			xmlWriter.writeComment("Event source file: " + path);

			// start the root element
			xmlWriter.writeCharacters("\n");
			xmlWriter.writeStartElement(ROOT_ELEMENT);
			xmlWriter.writeAttribute("numevents", "" + getEventCount());
            xmlWriter.writeCharacters("\n");

            // The difficulty is that this method can be called at
            // any time. So we need to save our state and then restore
            // it when we're done.
            ReaderState state = getState();

            // Go to the beginning
			rewind();

			// now loop through the events
			EvioEvent event;
			try {
				while ((event = parseNextEvent()) != null) {
					event.toXML(xmlWriter);
					// anybody interested in progress?
					if (progressListener != null) {
						progressListener.completed(event.getEventNumber(), getEventCount());
					}
				}
			}
			catch (EvioException e) {
				e.printStackTrace();
				return WriteStatus.UNKNOWN_ERROR;
			}

			// done. Close root element, end the document, and flush.
			xmlWriter.writeEndElement();
			xmlWriter.writeEndDocument();
			xmlWriter.flush();
			xmlWriter.close();

			try {
				fos.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}

            // Restore our original settings
            restoreState(state);

		}
		catch (XMLStreamException e) {
			e.printStackTrace();
			return WriteStatus.UNKNOWN_ERROR;
		}
        catch (FactoryConfigurationError e) {
            return WriteStatus.UNKNOWN_ERROR;
        }
        catch (EvioException e) {
            return WriteStatus.EVIO_EXCEPTION;
        }

		return WriteStatus.SUCCESS;
	}


    /**
     * This is the number of events in the file. Any dictionary event is <b>not</b>
     * included in the count. In versions 3 and earlier, it is not computed unless
     * asked for, and if asked for it is computed and cached.
     *
     * @return the number of events in the file.
     * @throws IOException if failed file access
     * @throws EvioException if read failure
     */
    public int getEventCount()  throws IOException, EvioException {

        if (evioVersion > 3) return eventCount;

        if (eventCount < 0) {
            // The difficulty is that this method can be called at
            // any time. So we need to save our state and then restore
            // it when we're done.
            ReaderState state = getState();

            rewind();
            eventCount = 0;

            while ((nextEvent()) != null) {
                eventCount++;
            }

            // Restore our original settings
            restoreState(state);
        }

        return eventCount;
    }

    /**
     * This is the number of blocks in the file including the empty
     * block usually at the end of version 4 files/buffers.
     * For version 3 files, a block size read from the first block is used
     * to calculate the result.
     * It is not computed unless in random access mode or is
     * asked for, and if asked for it is computed and cached.
     *
     * @return the number of blocks in the file (estimate for version 3 files)
     */
    public int getBlockCount() {

        if (evioVersion > 3) return blockCount;

        if (blockCount < 0) {
            // Although block size is theoretically adjustable, I believe
            // that everyone used 8192 words for the block size in version 3.
            blockCount = (int) (fileSize/firstBlockSize);
        }

        return blockCount;
    }


	/**
	 * Method used for diagnostics. It compares two event files. It checks the following (in order):<br>
	 * <ol>
	 * <li> That neither file is null.
	 * <li> That both files exist.
	 * <li> That neither file is a directory.
	 * <li> That both files can be read.
	 * <li> That both files are the same size.
	 * <li> That both files contain the same number of events.
	 * <li> Finally, that they are the same in a byte-by-byte comparison.
	 * </ol>
     * NOTE: Two files with the same events but different physical record size will be reported as different.
     *       They will fail the same size test.
     * @param evFile1 first file to be compared
     * @param evFile2 second file to be compared
	 * @return <code>true</code> if the files are, byte-by-byte, identical.
	 */
	public static boolean compareEventFiles(File evFile1, File evFile2) {

		if ((evFile1 == null) || (evFile2 == null)) {
			System.out.println("In compareEventFiles, one or both files are null.");
			return false;
		}

		if (!evFile1.exists() || !evFile2.exists()) {
			System.out.println("In compareEventFiles, one or both files do not exist.");
			return false;
		}

		if (evFile1.isDirectory() || evFile2.isDirectory()) {
			System.out.println("In compareEventFiles, one or both files is a directory.");
			return false;
		}

		if (!evFile1.canRead() || !evFile2.canRead()) {
			System.out.println("In compareEventFiles, one or both files cannot be read.");
			return false;
		}

		String name1 = evFile1.getName();
		String name2 = evFile2.getName();

		long size1 = evFile1.length();
		long size2 = evFile2.length();

		if (size1 == size2) {
			System.out.println(name1 + " and " + name2 + " have the same length: " + size1);
		}
		else {
			System.out.println(name1 + " and " + name2 + " have the different lengths.");
			System.out.println(name1 + ": " + size1);
			System.out.println(name2 + ": " + size2);
			return false;
		}

		try {
			EvioReader evioFile1 = new EvioReader(evFile1);
			EvioReader evioFile2 = new EvioReader(evFile2);
			int evCount1 = evioFile1.getEventCount();
			int evCount2 = evioFile2.getEventCount();
			if (evCount1 == evCount2) {
				System.out.println(name1 + " and " + name2 + " have the same #events: " + evCount1);
			}
			else {
				System.out.println(name1 + " and " + name2 + " have the different #events.");
				System.out.println(name1 + ": " + evCount1);
				System.out.println(name2 + ": " + evCount2);
				return false;
			}
		}
        catch (EvioException e) {
            e.printStackTrace();
            return false;
        }
        catch (IOException e) {
            e.printStackTrace();
            return false;
        }


		System.out.print("Byte by byte comparison...");
		System.out.flush();

		int onetenth = (int)(1 + size1/10);

		//now a byte-by-byte comparison
		try {
			FileInputStream fis1 = new FileInputStream(evFile1);
			FileInputStream fis2 = new FileInputStream(evFile1);

			for (int i = 0; i < size1; i++) {
				try {
					int byte1 = fis1.read();
					int byte2 = fis2.read();

					if (byte1 != byte2) {
						System.out.println(name1 + " and " + name2 + " different at byte offset: " + i);
						return false;
					}

					if ((i % onetenth) == 0) {
						System.out.print(".");
						System.out.flush();
					}
				}
				catch (IOException e) {
					e.printStackTrace();
					return false;
				}
			}

			System.out.println("");

			try {
				fis1.close();
				fis2.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		catch (FileNotFoundException e) {
			e.printStackTrace();
		}


		System.out.println("files " + name1 + " and " + evFile2.getPath() + " are identical.");
		return true;
	}

}